from __future__ import annotations

from collections.abc import Sequence
from types import NoneType
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    Primitives = int | float | str | bool | None


primitive_types: Final[Sequence[type[object]]] = [
    int,
    float,
    str,
    bool,
    type(None),
    NoneType,
]

primitive_type_mapping: Final[dict[type[object], str]] = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    type(None): "null",
    NoneType: "null",
}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "EYConf"
copyright = "2025, Sebastian B. Mohr"
author = "Sebastian B. Mohr"

master_doc = "index"
language = "en"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


templates_path = ["_templates"]
exclude_patterns = []


extensions = [
    "sphinx.ext.intersphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
    "sphinxcontrib.typer",
    "sphinx.ext.napoleon",
    "sphinx_design",
    # "myst_parser",
    "myst_nb",
]
autosummary_generate = True  # Turn on sphinx.ext.autosummary
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "jsonschema": ("https://python-jsonschema.readthedocs.io/en/stable", None),
}
nb_execution_mode = "off"
myst_enable_extensions = [
    "colon_fence",
    "deflist",
]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]
html_theme_options = {
    "light_logo": "logo_light.png",
    "dark_logo": "logo_dark.png",
    "light_css_variables": {
        "color-brand-primary": "#2f3992",
        "color-brand-content": "#dee2e6",
    },
    "dark_css_variables": {
        "color-brand-primary": "#2f3992",
        "color-brand-content": "#dee2e6",
    },
    # Sources for editing
    "source_view_link": "https://github.com/semohr/eyconf/blob/main/docs/{filename}",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/semohr/eyconf",
            "html": """
                <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round"><path d="M15 22v-4a4.8 4.8 0 0 0-1-3.5c3 0 6-2 6-5.5.08-1.25-.27-2.48-1-3.5.28-1.15.28-2.35 0-3.5 0 0-1 0-3 1.5-2.64-.5-5.36-.5-8 0C6 2 5 2 5 2c-.3 1.15-.3 2.35 0 3.5A5.403 5.403 0 0 0 4 9c0 3.5 3 5.5 6 5.5-.39.49-.68 1.05-.85 1.65-.17.6-.22 1.23-.15 1.85v4"/><path d="M9 18c-4.51 2-5-2-7-2"/></svg>
            """,
            "class": "",
        },
    ],
}
html_css_files = [
    "custom.css",
]



from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import asdict, dataclass, is_dataclass
from types import NoneType, UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Protocol,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    runtime_checkable,
)

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

log = logging.getLogger(__name__)

D = TypeVar("D", bound="DataclassInstance")


@dataclass
class AttributeDict:
    """A generic dataclass for holding dynamic attributes."""

    def __init__(self, **kwargs: Any):
        """Initialize the AttributeDict with given keyword arguments."""
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, name: str) -> Any:
        """Get attribute dynamically. If it does not exist, we create it."""
        if name.startswith("_"):
            raise AttributeError(f"{name} not found")

        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            n = AttributeDict()
            setattr(self, name, n)
            return n

    def __setattr__(self, name: str, value: Any):
        """Set attribute dynamically."""
        if isinstance(value, dict):
            value = AttributeDict(**value)
        object.__setattr__(self, name, value)

    def __getitem__(self, key: str) -> Any:
        """Get item dynamically."""
        return self.__getattr__(key)

    def as_dict(self) -> dict:
        """Convert the AttributeDict to a standard dictionary."""
        result = {}
        for key, value in self.__dict__.items():
            if isinstance(value, AttributeDict):
                result[key] = value.as_dict()
            else:
                result[key] = value
        return result

    def __deepcopy__(self, memo: dict) -> AttributeDict:
        """Create a deep copy of the AttributeDict."""
        # Avoid infinite recursion with memo
        if id(self) in memo:
            return memo[id(self)]

        # Create new instance
        new_instance = AttributeDict()
        memo[id(self)] = new_instance

        # Deep copy all attributes
        for key, value in self.__dict__.items():
            # Use copy.deepcopy for nested objects, but handle AttributeDict specially
            if isinstance(value, AttributeDict):
                setattr(new_instance, key, deepcopy(value, memo))
            else:
                setattr(new_instance, key, deepcopy(value, memo))

        return new_instance

    def __repr__(self) -> str:
        """Representation of the AttributeDict."""
        return f"AttributeDict({self.as_dict()})"

    def __str__(self) -> str:
        """Use the dict string representation."""
        return str(self.as_dict())

    def __bool__(self) -> bool:
        """Return False if the AttributeDict is empty, True otherwise."""
        return bool(self.__dict__)


class AccessProxy(Generic[D]):
    """Proxy to access attributes dynamically."""

    _data: D
    _extra_data: AttributeDict

    def __init__(self, data: D, extra_data: AttributeDict):
        self._data = data
        self._extra_data = extra_data

    def __getattr__(self, name: str) -> Any:
        """Get attribute from either the typed data or additional data."""
        try:
            ret = getattr(self._data, name)
            # We need to wrap nested dataclasses as well
            # Needed for accessing a mixed case, where we add an unknown property to
            # a nested schema. In this case, we need the same extra level in _extra_data.
            if is_dataclass(ret):
                return AccessProxy(ret, getattr(self._extra_data, name))  # type: ignore[arg-type]
            return getattr(self._data, name)
        except AttributeError:
            return getattr(self._extra_data, name)

    def __getitem__(self, key: str) -> Any:
        """Get item dynamically."""
        return self.__getattr__(key)

    def __setattr__(self, name: str, value: Any):
        """Set attribute on either the typed data or additional data."""
        if name.startswith("_"):
            object.__setattr__(self, name, value)
        else:
            if hasattr(self._data, name):
                setattr(self._data, name, value)
            else:
                setattr(self._extra_data, name, value)

    def __delattr__(self, name: str):
        """Delete attribute from either the typed data or additional data."""
        if name.startswith("_"):
            object.__delattr__(self, name)
        else:
            if hasattr(self._data, name):
                delattr(self._data, name)
            else:
                delattr(self._extra_data, name)

    def as_dict(self) -> dict:
        """Convert the AccessProxy to a standard dictionary."""
        merged = deepcopy(self._extra_data.as_dict())
        data_dict = deepcopy(asdict(self._data))
        result = merge_dicts(data_dict, merged)
        return result


def merge_dicts(a: dict, b: dict, path=[]):
    """Merge dict b into dict a, raising an exception on conflicts."""
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key], path + [str(key)])
            elif a[key] != b[key]:
                raise Exception("Conflict at " + ".".join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def dataclass_from_dict(in_type: type[D], data: dict) -> D:
    """Convert a dict to a dataclass instance of the given type. Always returns a dataclass."""
    result = _dataclass_from_dict_inner(in_type, data)
    if result is None:
        raise ValueError(f"Could not parse data {data} with type {in_type}")
    return result


def _dataclass_from_dict_inner(target_type: type, data: Any) -> Any:
    """Inner function that handles Union types and may return None."""
    # Handle Union types
    origin = get_origin(target_type)
    if origin is Union or origin is UnionType:
        args = get_args(target_type)
        includes_none = any(arg is NoneType or arg is type(None) for arg in args)

        if data is None and includes_none:
            return None

        for arg in args:
            if arg is NoneType or arg is type(None):
                continue
            try:
                return _dataclass_from_dict_inner(arg, data)
            except (ValueError, TypeError, KeyError):
                continue
        return None

    # Handle dict data - convert to dataclass
    if isinstance(data, dict) and is_dataclass(target_type):
        field_types = get_type_hints(target_type, include_extras=False)

        found_fields = {}
        for field_name, field_type in field_types.items():
            if field_name in data:
                found_fields[field_name] = _dataclass_from_dict_inner(
                    field_type, data[field_name]
                )

        try:
            return target_type(**found_fields)  # type: ignore[bad-instantiation]
        except TypeError as e:
            raise ValueError(f"Failed to create {target_type.__name__}: {e}")

    # Potentially nested dataclass in dicts
    if isinstance(data, dict) and get_origin(target_type) is dict:
        key_type, value_type = get_args(target_type)
        return {
            _dataclass_from_dict_inner(key_type, k): _dataclass_from_dict_inner(
                value_type, v
            )
            for k, v in data.items()
        }

    # Handle sequence types (list, tuple)
    if isinstance(data, (list, tuple)):
        if hasattr(target_type, "__args__") and target_type.__args__:
            elem_type = target_type.__args__[0]
            return [_dataclass_from_dict_inner(elem_type, item) for item in data]
        else:
            return data

    # Handle primitive types
    return data


T = TypeVar("T")


@runtime_checkable
class DictAccess(Protocol):
    """Protocol for dict-like access."""

    def __getitem__(self, key: str) -> Any: ...  # noqa: D105


def dict_access(cls: type[T]) -> type[T]:
    """Class decorator to add dict-like access to class attributes.

    Can be used to add `dict`-like access to any class, allowing
    attribute access via the `obj['attribute']` syntax.

    Use with care, dict-style access does not provide any type safety
    and will not be checked by static type checkers.

    Usage:

    ```python
    @dict_access
    class MySchema:
        forty_two: int = 42

    obj = MySchema()
    assert isinstance(obj, DictAccess)
    print(obj['forty_two'])  # Outputs: 42
    ```
    """

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    setattr(cls, "__getitem__", __getitem__)
    return cls



from .config import EYConf, EYConfBase, EYConfExtraFields
from .generate_yaml import dataclass_to_yaml
from .validation import validate

__all__ = [
    "dataclass_to_yaml",
    "EYConf",
    "EYConfBase",
    "EYConfExtraFields",
    "validate",
]



from collections.abc import Sequence
from collections.abc import Sequence as ABCSequence
from dataclasses import is_dataclass
from functools import cache
from types import NoneType, UnionType
from typing import Annotated, Any, Literal, Union, get_args, get_origin

from jsonschema import Draft202012Validator
from typing_extensions import NotRequired

from eyconf.constants import primitive_type_mapping
from eyconf.type_utils import get_type_hints_resolve_namespace

SchemaType = dict[str, Any] | dict[str, str] | dict[Any, Any]


__all__ = ["to_json_schema"]


@cache
def to_json_schema(
    type: type,
    check_schema: bool = True,
    allow_additional: bool = True,
) -> dict:
    """Convert a TypedDict or dataclass to a JSON schema.

    Parameters
    ----------
    type : type
        The TypedDict or dataclass to convert.
    check_schema : bool
        Whether to check the schema for validity.
    allow_additional : bool
        Whether to allow extra, unrecognized properties in the schema.

    Raises
    ------
    ValueError
        If the type is not supported.
    SchemaError
        If the schema is invalid (should not happen).
    """
    schema: dict[str, Any] = {
        "type": "object",
        "properties": {},
        "required": [],
        "additionalProperties": allow_additional,
    }

    # Get type hints for the TypedDict
    type_hints = get_type_hints_resolve_namespace(type, include_extras=True)

    # Add the type hints to the schema
    for field_name, field_type in type_hints.items():
        p, r = __convert_type_to_schema(field_type, allow_additional=allow_additional)
        schema["properties"][field_name] = p

        if r:
            schema["required"].append(field_name)

    if check_schema:
        Draft202012Validator.check_schema(schema)

    return schema


def __convert_type_to_schema(
    field_type: type,
    **kwargs,
) -> tuple[SchemaType, bool]:
    """
    Convert a type to a JSON schema.

    Parameters
    ----------
    field_type : type
        The type to convert.
    **kwargs : dict
        Passed recusrively and to `to_json_schema()` if called in here.

    Returns
    -------
    schema : dict | None
        The JSON schema. None if the type is not supported or is is resolves to optional none value.
    is_required : bool
        Whether the field is required.
    """
    is_required = True

    origin = get_origin(field_type)

    # Unpack annotated types
    if origin is Annotated:
        # We always assume the first argument is the type
        # all other arguments are metadata (docstrings)
        field_type = get_args(field_type)[0]
        return __convert_type_to_schema(field_type, **kwargs)

    # Handle Literal
    if origin is Literal:
        allowed_values = get_args(field_type)
        return {
            "type": __infer_type_from_values(allowed_values),
            "enum": list(allowed_values),
        }, is_required

    # Handle NotRequired
    if origin is NotRequired:
        field_type = get_args(field_type)[0]
        is_required = False
        return __convert_type_to_schema(field_type, **kwargs)[0], is_required

    # Handler union
    if origin is Union or origin is UnionType:
        allowed_types = set(get_args(field_type))
        if NoneType in allowed_types:
            is_required = False
            allowed_types.remove(NoneType)

        if len(allowed_types) > 1:
            return {
                "anyOf": [
                    __convert_type_to_schema(allowed_type, **kwargs)[0]
                    for allowed_type in allowed_types
                ]
            }, is_required
        else:
            # case length == 1
            # Another case is impossible as Union must have at least one type and
            # get simplified beforehand if it does not Union[int] -> int
            t, _ = __convert_type_to_schema(allowed_types.pop(), **kwargs)
            return t, is_required

    # Handle sequence types
    if origin in [list, set, tuple, Sequence, ABCSequence]:
        return {
            "type": "array",
            "items": __convert_type_to_schema(get_args(field_type)[0], **kwargs)[0],
        }, is_required

    # Handle TypedDict and dataclasses
    try:
        if issubclass(field_type, dict) or is_dataclass(field_type):
            return to_json_schema(field_type, **kwargs), is_required
    except TypeError:
        # Throws an error in case of a type that is not a class
        pass

    # Handle Dicts - arbitrary keys with typed values
    if origin in [dict, dict]:
        key_type, value_type = get_args(field_type)
        if key_type is not str:
            raise ValueError("Only string keys are supported in dict types")

        return {
            "type": "object",
            "patternProperties": {
                ".*": __convert_type_to_schema(value_type, **kwargs)[0]
            },
        }, is_required

    # Handle other types
    match = primitive_type_mapping.get(field_type)

    if match:
        return {"type": match}, is_required

    if field_type is Any:
        return {}, is_required

    raise ValueError(f"Unsupported type: {field_type}")


def __infer_type_from_values(values: tuple | list):
    types: list[type] = []
    for value in values:
        value_type = type(value)
        if value_type not in types:
            types.append(value_type)

    type_names: list[str] = []
    for t in types:
        if t in primitive_type_mapping:
            type_names.append(primitive_type_mapping[t])
        else:
            raise ValueError(f"Unsupported literal type: {t}")
    if len(type_names) == 1:
        return type_names[0]
    else:
        return type_names



from __future__ import annotations

from dataclasses import dataclass, field

import pytest
from eyconf import EYConfBase

from eyconf.utils import AttributeDict
from eyconf.validation import MultiConfigurationError


@dataclass
class Config42:
    int_field: int = 42
    str_field: str = "FortyTwo!"


@dataclass
class ConfigNested:
    nested: Config42 = field(default_factory=Config42)
    nested_optional: Config42 | None = None
    other_field: str = "Hello, World!"


@pytest.fixture
def conf42() -> EYConfBase[Config42]:
    return EYConfBase(Config42(), schema=Config42)


@pytest.fixture
def conf_nested() -> EYConfBase[ConfigNested]:
    return EYConfBase(ConfigNested(), schema=ConfigNested)


class TestCreation:
    def test_init(self):
        conf = EYConfBase(Config42(), schema=Config42)
        assert conf.data.int_field == 42
        assert conf.data.str_field == "FortyTwo!"
        assert isinstance(conf.data, Config42)

    def test_init_dict(self):
        conf_dict = EYConfBase(
            {"int_field": 100, "str_field": "Dict value!"},
            schema=Config42,
        )
        assert conf_dict.data.int_field == 100
        assert conf_dict.data.str_field == "Dict value!"
        assert isinstance(conf_dict.data, Config42)

    def test_init_dict_no_schema(self):
        conf = EYConfBase(Config42())
        assert conf.data.int_field == 42
        assert conf.data.str_field == "FortyTwo!"
        assert isinstance(conf.data, Config42)

        with pytest.raises(ValueError):
            EYConfBase(
                {"int_field": "not_an_int", "str_field": "Dict value!"},
            )

    def test_init_with_invalid_data(self):
        with pytest.raises(MultiConfigurationError):
            EYConfBase({"int_field": "not_an_int"}, schema=Config42)

    def test_init_with_missing_required_fields(self):
        with pytest.raises(MultiConfigurationError):
            EYConfBase({"str_field": "test"}, schema=Config42)

    def test_init_with_extra_fields(
        self,
    ):
        data = {"int_field": 42, "str_field": "test", "extra_field": "unexpected"}
        with pytest.raises(MultiConfigurationError):
            EYConfBase(
                data,
                schema=Config42,
            )

        conf = EYConfBase(
            data,
            schema=Config42,
            allow_additional_properties=True,
        )
        assert conf.data.int_field == 42
        assert conf.data.str_field == "test"
        assert conf.data.extra_field == "unexpected"  # type: ignore


class TestUpdate:
    """Update should allow to apply partial updates to the configuration,
    validating only the provided fields, and leaving others unchanged.
    """

    def test_simple(self, conf42: EYConfBase[Config42]):
        conf42.update({"int_field": 100, "str_field": "Updated value!"})

        assert conf42.data.int_field == 100
        assert conf42.data.str_field == "Updated value!"

    def test_partial(self, conf42: EYConfBase[Config42]):
        conf42.update({"int_field": 100})

        assert conf42.data.int_field == 100
        assert conf42.data.str_field == "FortyTwo!"

    def test_nested(self, conf_nested: EYConfBase[ConfigNested]):
        conf_nested.update(
            {
                "nested": {"int_field": 100, "str_field": "Updated nested value!"},
                "nested_optional": {"int_field": 200, "str_field": "Optional nested!"},
                "other_field": "Updated parent value!",
            }
        )

        assert conf_nested.data.nested.int_field == 100
        assert conf_nested.data.nested.str_field == "Updated nested value!"
        assert conf_nested.data.nested_optional.int_field == 200  # type: ignore[union-attr]
        assert conf_nested.data.nested_optional.str_field == "Optional nested!"  # type: ignore[union-attr]
        assert conf_nested.data.other_field == "Updated parent value!"

    def test_nested_partial(self, conf_nested: EYConfBase[ConfigNested]):
        conf_nested.update(
            {
                "nested": {"int_field": 100},
            }
        )

        assert conf_nested.data.nested.int_field == 100
        assert conf_nested.data.nested.str_field == "FortyTwo!"
        assert conf_nested.data.other_field == "Hello, World!"
        assert conf_nested.data.nested_optional is None

    def test_invalid(self, conf42: EYConfBase[Config42]):
        with pytest.raises(MultiConfigurationError):
            conf42.update({"int_field": "not an int"})

        assert conf42.data.int_field == 42

    def test_additional_fields(self, conf42: EYConfBase[Config42]):
        conf42.allow_additional_properties = True
        conf42.update({"int_field": 100, "new_field": "I am new!"})

        assert conf42.data.int_field == 100
        assert conf42.data.new_field == "I am new!"  # type: ignore[attr-defined]

    def test_additional_fields_nested(self, conf_nested: EYConfBase[ConfigNested]):
        conf_nested.allow_additional_properties = True
        conf_nested.update(
            {
                "nested": {"new_nested_field": {"foo": "bar"}},
            }
        )

        assert conf_nested.data.nested.new_nested_field.foo == "bar"  # type: ignore[attr-defined]
        assert isinstance(conf_nested.data.nested.new_nested_field, AttributeDict)  # type: ignore[attr-defined]

    def test_additional_fields_nested_deep(self, conf_nested: EYConfBase[ConfigNested]):
        conf_nested.allow_additional_properties = True
        # Even more nesting
        conf_nested.update(
            {
                "nested": {"d1": {"d2": {"d3": {"d4_field": "deep value!"}}}},
            }
        )
        assert conf_nested.data.nested.d1.d2.d3.d4_field == "deep value!"  # type: ignore[attr-defined]


class TestOverwrite:
    def test_simple(self, conf42: EYConfBase[Config42]):
        conf42.overwrite(
            {
                "int_field": 100,
                "str_field": "Overwritten value!",
            }
        )

        assert conf42.data.int_field == 100
        assert conf42.data.str_field == "Overwritten value!"

        # Overwrite with Config instance
        new_conf = Config42(int_field=200, str_field="Another overwrite!")
        conf42.overwrite(new_conf)
        assert conf42.data.int_field == 200
        assert conf42.data.str_field == "Another overwrite!"

    def test_invalid(self, conf42: EYConfBase[Config42]):
        with pytest.raises(MultiConfigurationError):
            conf42.overwrite({"int_field": "not an int", "str_field": "Valid str"})

        with pytest.raises(MultiConfigurationError):
            conf42.overwrite({"str_field": "Missing int field"})

    def test_nested(self, conf_nested: EYConfBase[ConfigNested]):
        conf_nested.overwrite(
            {
                "nested": {"int_field": 100, "str_field": "Overwritten nested!"},
                "nested_optional": {"int_field": 200, "str_field": "Optional nested!"},
                "other_field": "Overwritten parent!",
            }
        )

        assert conf_nested.data.nested.int_field == 100
        assert conf_nested.data.nested.str_field == "Overwritten nested!"
        assert conf_nested.data.nested_optional.int_field == 200  # type: ignore[union-attr]
        assert conf_nested.data.nested_optional.str_field == "Optional nested!"  # type: ignore[union-attr]
        assert conf_nested.data.other_field == "Overwritten parent!"


class TestReset:
    def test_simple(self, conf42: EYConfBase[Config42]):
        conf42.update(
            {
                "int_field": 100,
                "str_field": "Updated value!",
            }
        )

        assert conf42.data.int_field == 100
        assert conf42.data.str_field == "Updated value!"

        conf42.reset()

        assert conf42.data.int_field == 42
        assert conf42.data.str_field == "FortyTwo!"

    def test_nested(self, conf_nested: EYConfBase[ConfigNested]):
        conf_nested.allow_additional_properties = True
        conf_nested.update(
            {
                "nested": {"int_field": 100, "str_field": "Updated nested value!"},
                "nested_optional": {"int_field": 200, "str_field": "Optional nested!"},
                "other_field": "Updated parent value!",
                "extra_field": "I am extra!",  # type: ignore[attr-defined]
            }
        )

        assert conf_nested.data.nested.int_field == 100
        assert conf_nested.data.nested.str_field == "Updated nested value!"
        assert conf_nested.data.nested_optional.int_field == 200  # type: ignore[union-attr]
        assert conf_nested.data.nested_optional.str_field == "Optional nested!"  # type: ignore[union-attr]
        assert conf_nested.data.other_field == "Updated parent value!"

        conf_nested.reset()

        assert conf_nested.data.nested.int_field == 42
        assert conf_nested.data.nested.str_field == "FortyTwo!"
        assert conf_nested.data.nested_optional is None
        assert conf_nested.data.other_field == "Hello, World!"
        assert not hasattr(conf_nested.data, "extra_field")


class TestConverters:
    def test_schema_data_as_dict(self, conf42: EYConfBase[Config42]):
        expected = {
            "int_field": 42,
            "str_field": "FortyTwo!",
        }
        assert conf42.to_dict() == expected

    def test_extra_data_as_dict(self, conf42: EYConfBase[Config42]):
        conf42.allow_additional_properties = True
        conf42.update({"new_field": "I am new!"})

        assert conf42.to_dict(include_additional=True) == {
            "int_field": 42,
            "str_field": "FortyTwo!",
            "new_field": "I am new!",
        }
        assert conf42.to_dict() == {
            "int_field": 42,
            "str_field": "FortyTwo!",
        }

    def test_extra_data_as_dict_nested(self, conf_nested: EYConfBase[ConfigNested]):
        conf_nested.allow_additional_properties = True
        conf_nested.update(
            {
                "nested": {"new_nested_field": {"foo": "bar"}},
                "another_extra": 123,
            }
        )

        assert conf_nested.to_dict(include_additional=True) == {
            "nested": {
                "int_field": 42,
                "str_field": "FortyTwo!",
                "new_nested_field": {"foo": "bar"},
            },
            "nested_optional": None,
            "other_field": "Hello, World!",
            "another_extra": 123,
        }
        assert conf_nested.to_dict() == {
            "nested": {
                "int_field": 42,
                "str_field": "FortyTwo!",
            },
            "nested_optional": None,
            "other_field": "Hello, World!",
        }

    def test_to_yaml(self, conf42: EYConfBase[Config42]):
        expected_yaml = "int_field: 42\nstr_field: FortyTwo!"
        assert conf42.to_yaml() == expected_yaml

    def test_update_additional_property_list(self, conf42: EYConfBase[Config42]):
        """Test updating additional list properties"""
        conf42.allow_additional_properties = True

        conf42.update({"extra_list": [1, 2, 3]})
        conf42.update({"extra_list": ["a", "b", "c"]})

        assert conf42.data.extra_list == ["a", "b", "c"]  # type: ignore[attr-defined]

    def test_error_without_additional_properties(self, conf42: EYConfBase[Config42]):
        """Test that updating additional properties without allowing them raises an error"""
        with pytest.raises(AttributeError):
            conf42.update({"extra_field": "not allowed"})


class TestPrintingUtils:
    @pytest.mark.parametrize(
        "fixture_name,expected_lines",
        [
            ("conf42", ["  int_field: 42", "  str_field: FortyTwo!"]),
            (
                "conf_nested",
                ["  nested:", "    int_field: 42", "  other_field: Hello, World!"],
            ),
        ],
    )
    def test_str_formats_data(self, fixture_name, expected_lines, request):
        """Test that __str__ properly formats configuration data"""
        conf = request.getfixturevalue(fixture_name)
        str_output = str(conf)

        for expected_line in expected_lines:
            assert expected_line in str_output

    def test_repr_includes_object_info_and_data(self, conf42: EYConfBase[Config42]):
        """Test that __repr__ includes basic object info and formatted data"""
        repr_str = repr(conf42)

        # Basic object info
        assert repr_str.startswith("<EYConfBase object at 0x")
        # Delegates to __str__ for data
        assert "int_field: 42" in repr_str

    @pytest.mark.parametrize(
        "test_data,expected",
        [
            ({"key": "value"}, ["key: value"]),
            ({"nested": {"inner": "val"}}, ["nested:", "    inner: val"]),
            ({}, []),
        ],
    )
    def test_pretty_format_basic_cases(
        self, test_data, expected, conf42: EYConfBase[Config42]
    ):
        """Test _pretty_format with basic cases"""
        formatted = conf42._pretty_format(test_data)

        if expected:
            for line in expected:
                assert line in formatted
        else:
            assert formatted == ""



from eyconf.utils import AttributeDict, AccessProxy, dict_access, DictAccess
from dataclasses import dataclass, field


@dict_access
@dataclass
class Nested:
    str_field: str = "FortyTwo"


@dict_access
@dataclass
class Config42:
    int_field: int = 42
    nested: Nested = field(default_factory=Nested)


class TestAttributeDict:
    def test_init(self):
        attr_dict = AttributeDict(**{"foo": "bar", "nested": {"level": 42}})
        assert isinstance(attr_dict, AttributeDict)
        assert attr_dict.foo == "bar"

        assert isinstance(attr_dict.nested, AttributeDict)
        assert attr_dict.nested.level == 42

    def test_set_get(self):
        attr_dict = AttributeDict()
        attr_dict.foo = "bar"  # type: ignore

        assert attr_dict.foo == "bar"

    def test_nested(self):
        attr_dict = AttributeDict()
        attr_dict.nested.level = 42

        assert attr_dict.nested.level == 42

    def test_as_dict(self):
        attr_dict = AttributeDict()
        attr_dict.foo = "bar"  # type: ignore
        attr_dict.nested.level = 42  # type: ignore

        expected = {
            "foo": "bar",
            "nested": {
                "level": 42,
            },
        }

        assert attr_dict.as_dict() == expected

    def test_bool_conversion(self):
        attr_dict = AttributeDict()
        attr_dict.foo = "bar"  # type: ignore

        assert bool(attr_dict) is True

        empty_attr_dict = AttributeDict()
        assert bool(empty_attr_dict) is False


class TestAccessProxy:
    def test_access_fails(self):
        config_data = Config42()
        extra_data = AttributeDict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )

        assert proxy.int_field == 42

    def test_set(self):
        config_data = Config42()
        extra_data = AttributeDict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )
        proxy.int_field = 100

        assert proxy.int_field == 100
        assert config_data.int_field == 100

    def test_del(self):
        config_data = Config42()
        extra_data = AttributeDict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )

        assert proxy.int_field == 42

        del proxy.int_field
        del proxy._data

    def test_set_extra(self):
        config_data = Config42()
        extra_data = AttributeDict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )
        proxy.foo = "bar"  # type: ignore

        assert proxy.int_field == 42
        assert proxy.foo == "bar"

        # make sure it actually sits in _additional_data
        assert proxy._extra_data.foo == "bar"  # type: ignore

    def test_del_extra(self):
        config_data = Config42()
        extra_data = AttributeDict()
        proxy = AccessProxy(
            config_data,
            extra_data,
        )
        proxy.foo = "bar"  # type: ignore

        assert proxy.foo == "bar"

        del proxy.foo  # type: ignore


class TestDictStyleAccess:
    """Test the @dict_access decorator functionality."""

    def test_get_data(self):
        config = Config42()

        # Note that placing `assert isinstance(config, DictAccess)` here
        # will make some runtime type checkers error about the attribute access
        # Once you place the assert, better stick to dict-style access
        # for the rest of the scope
        assert isinstance(config, DictAccess)
        assert isinstance(config["nested"], DictAccess)

        assert config.int_field == 42
        assert config.nested.str_field == "FortyTwo"

        assert config["int_field"] == 42
        assert config["nested"]["str_field"] == "FortyTwo"



from __future__ import annotations
from dataclasses import fields
from pprint import pprint
from typing import (
    Any,
    Literal,
    Optional,
    TypedDict,
)
from collections.abc import Sequence
from typing_extensions import NotRequired
import pytest
from eyconf.type_utils import get_type_hints_resolve_namespace
from eyconf.validation import to_json_schema
from dataclasses import dataclass

from typing import Annotated


class TestToSchema:
    """
    Test the dataclass to json schema function
    """

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_primitives(self, as_dataclass, allow_additional):
        """Test a simple dataclass."""

        @dataclass
        class Primitives:
            foo: str
            bar: int
            baz: float
            qux: bool
            nay: None

        if not as_dataclass:
            Primitives = dataclass_to_typeddict(Primitives)  # type: ignore

        schema = to_json_schema(Primitives, allow_additional=allow_additional)

        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string"},
                "bar": {"type": "integer"},
                "baz": {"type": "number"},
                "qux": {"type": "boolean"},
                "nay": {"type": "null"},
            },
            "required": ["foo", "bar", "baz", "qux", "nay"],
            "additionalProperties": allow_additional,
        }

        @dataclass
        class InvalidPrimitive:
            foo: bytes

        if not as_dataclass:
            InvalidPrimitive = dataclass_to_typeddict(InvalidPrimitive)  # type: ignore

        with pytest.raises(ValueError):
            to_json_schema(InvalidPrimitive)

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_literal(self, as_dataclass, allow_additional):
        @dataclass
        class Schema:
            foo: Literal["bar", "baz"]
            bar: Literal[None]

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema)  # type: ignore

        schema = to_json_schema(Schema, allow_additional=allow_additional)

        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string", "enum": ["bar", "baz"]},
                "bar": {"type": "null", "enum": [None]},
            },
            "required": ["foo", "bar"],
            "additionalProperties": allow_additional,
        }

        @dataclass
        class InvalidLiteral:
            foo: Literal[b"0"]

        if not as_dataclass:
            InvalidLiteral = dataclass_to_typeddict(InvalidLiteral)  # type: ignore

        with pytest.raises(ValueError):
            to_json_schema(InvalidLiteral)

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_literal_with_different_types(self, as_dataclass, allow_additional):
        @dataclass
        class Schema:
            foo: Literal["a", "b", False]
            bar: Literal[1, None]

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema, allow_additional=allow_additional)  # type: ignore

        schema = to_json_schema(Schema, allow_additional=allow_additional)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": ["string", "boolean"], "enum": ["a", "b", False]},
                "bar": {"type": ["integer", "null"], "enum": [1, None]},
            },
            "required": ["foo", "bar"],
            "additionalProperties": allow_additional,
        }

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_optional(self, as_dataclass, allow_additional):
        @dataclass
        class Schema:
            foo: str | None
            bar: int | None
            baz: float

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema, allow_additional=allow_additional)  # type: ignore

        schema = to_json_schema(Schema, allow_additional=allow_additional)
        print(schema)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string"},
                "bar": {"type": "integer"},
                "baz": {"type": "number"},
            },
            "required": ["baz"],
            "additionalProperties": allow_additional,
        }

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_union(self, as_dataclass, allow_additional):
        @dataclass
        class Schema:
            foo: str | int
            bar: int | float

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema, allow_additional=allow_additional)  # type: ignore

        schema = to_json_schema(Schema, allow_additional=allow_additional)

        assert sorted(
            schema["properties"]["foo"]["anyOf"], key=lambda x: x["type"]
        ) == [{"type": "integer"}, {"type": "string"}]
        assert sorted(
            schema["properties"]["bar"]["anyOf"], key=lambda x: x["type"]
        ) == [{"type": "integer"}, {"type": "number"}]

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_nested_dict(self, as_dataclass, allow_additional):
        @dataclass
        class Dict1:
            foo: str

        if not as_dataclass:
            Dict1 = dataclass_to_typeddict(Dict1)  # type: ignore

        @dataclass
        class NestedTyped:
            dict1: Dict1
            dict_opt: Dict1 | None
            dict_uni: Dict1 | str
            baz: float | None

        if not as_dataclass:
            NestedTyped = dataclass_to_typeddict(NestedTyped)  # type: ignore

        schema = to_json_schema(NestedTyped, allow_additional=allow_additional)
        pprint(schema)
        assert schema["type"] == "object"
        assert schema["required"] == ["dict1", "dict_uni"]
        assert ["dict1", "dict_opt", "dict_uni", "baz"] == list(
            schema["properties"].keys()
        )

        dict1_obj = {
            "type": "object",
            "properties": {"foo": {"type": "string"}},
            "required": ["foo"],
            "additionalProperties": allow_additional,
        }

        assert schema["properties"]["dict1"] == dict1_obj
        assert schema["properties"]["dict_opt"] == dict1_obj

        assert "anyOf" in schema["properties"]["dict_uni"]
        assert (
            schema["properties"]["dict_uni"]["anyOf"][0] == dict1_obj
            or schema["properties"]["dict_uni"]["anyOf"][1] == dict1_obj
        )

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_lists(self, as_dataclass, allow_additional):
        @dataclass
        class Schema:
            foo: list[str]
            bar: Sequence[int] | None

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema)  # type: ignore

        schema = to_json_schema(Schema, allow_additional=allow_additional)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "array", "items": {"type": "string"}},
                "bar": {"type": "array", "items": {"type": "integer"}},
            },
            "required": [
                "foo",
            ],
            "additionalProperties": allow_additional,
        }

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_dicts(self, as_dataclass, allow_additional):
        @dataclass
        class Schema:
            foo: dict[str, int]
            bar: dict[str, str]

        if not as_dataclass:
            Schema = dataclass_to_typeddict(Schema)  # type: ignore

        schema = to_json_schema(Schema, allow_additional=allow_additional)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {
                    "type": "object",
                    "patternProperties": {
                        ".*": {"type": "integer"},
                    },
                },
                "bar": {
                    "type": "object",
                    "patternProperties": {
                        ".*": {"type": "string"},
                    },
                },
            },
            "required": ["foo", "bar"],
            "additionalProperties": allow_additional,
        }

    @pytest.mark.parametrize(
        ["as_dataclass", "allow_additional"],
        [(True, False), (True, False)],
    )
    def test_dict_nested(self, as_dataclass, allow_additional):
        @dataclass
        class Inner:
            inner: int

        @dataclass
        class Outer:
            outer: dict[str, Inner]

        if not as_dataclass:
            Outer = dataclass_to_typeddict(Outer)  # type: ignore

        schema = to_json_schema(Outer, allow_additional=allow_additional)

        assert schema == {
            "type": "object",
            "properties": {
                "outer": {
                    "type": "object",
                    "patternProperties": {
                        ".*": {
                            "type": "object",
                            "properties": {
                                "inner": {"type": "integer"},
                            },
                            "required": ["inner"],
                            "additionalProperties": allow_additional,
                        }
                    },
                },
            },
            "required": ["outer"],
            "additionalProperties": allow_additional,
        }

    def test_not_required(self):
        class MyTypedDict1(TypedDict):
            foo: NotRequired[str]
            bar: NotRequired[int]
            baz: NotRequired[float]
            qux: NotRequired[bool]
            nay: NotRequired[None]

        schema = to_json_schema(MyTypedDict1)

        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string"},
                "bar": {"type": "integer"},
                "baz": {"type": "number"},
                "qux": {"type": "boolean"},
                "nay": {"type": "null"},
            },
            "required": [],
            "additionalProperties": True,
        }

    def test_special(self):
        @dataclass
        class Schema:
            foo: None

        schema = to_json_schema(Schema)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "null"},
            },
            "required": ["foo"],
            "additionalProperties": True,
        }

        @dataclass
        class SchemaAny:
            foo: Any

        schema = to_json_schema(SchemaAny)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {},
            },
            "required": ["foo"],
            "additionalProperties": True,
        }

        @dataclass
        class UnionNone:
            foo: Optional[None]  # noqa: UP045

        schema = to_json_schema(UnionNone)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "null"},
            },
            "required": ["foo"],
            "additionalProperties": True,
        }

    def test_cache_hit(self):
        @dataclass
        class Schema:
            foo: str

        schema = to_json_schema(Schema)
        schema2 = to_json_schema(Schema)

        assert schema is schema2

    def test_non_str_dict(self):
        @dataclass
        class Schema:
            foo: dict[int, str]

        with pytest.raises(
            ValueError,
            match="Only string keys are supported in dict types",
        ):
            to_json_schema(Schema)

    def test_annotated(self):
        @dataclass
        class Schema:
            foo: Annotated[str, "some metadata"]
            bar: Annotated[int, "some metadata", "more metadata"]

        schema = to_json_schema(Schema)
        assert schema == {
            "type": "object",
            "properties": {
                "foo": {"type": "string"},
                "bar": {"type": "integer"},
            },
            "required": ["foo", "bar"],
            "additionalProperties": True,
        }

        @dataclass
        class NestedAnnotated:
            schema: Schema

        schema = to_json_schema(NestedAnnotated)
        assert schema == {
            "type": "object",
            "properties": {
                "schema": {
                    "type": "object",
                    "properties": {
                        "foo": {"type": "string"},
                        "bar": {"type": "integer"},
                    },
                    "required": ["foo", "bar"],
                    "additionalProperties": True,
                }
            },
            "required": ["schema"],
            "additionalProperties": True,
        }


# This function converts a dataclass to a TypedDict
def dataclass_to_typeddict(dc_cls: type):
    """Convert a dataclass to a TypedDict."""
    # Fetch type hints of the dataclass
    type_hints = get_type_hints_resolve_namespace(dc_cls)

    # Extract the fields and their types
    typeddict_fields: dict = {
        field.name: type_hints[field.name] for field in fields(dc_cls)
    }

    # Create the TypedDict dynamically
    return TypedDict(f"{dc_cls.__name__}Dict", typeddict_fields)  # type: ignore



"""Validating configuration data against schemas."""

from __future__ import annotations

import logging
from dataclasses import asdict, is_dataclass
from typing import TYPE_CHECKING, TypeVar, cast

from eyconf.utils import dataclass_from_dict

log = logging.getLogger(__name__)

from jsonschema import Draft202012Validator, ValidationError

from ._to_json import to_json_schema

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

__all__ = [
    "to_json_schema",
    "validate_json",
    "ConfigurationError",
    "MultiConfigurationError",
]

D = TypeVar("D", bound="DataclassInstance")


def validate(data: D | dict, schema: type[D]) -> D:
    """Validate the provided data against the schema and return the dataclass instance.

    This function first converts the schema dataclass to a JSON schema using
    `to_json_schema`, then validates the provided data against this schema
    using `validate_json`. If validation is successful, it returns an instance
    of the schema dataclass populated with the validated data.

    For more controls over validation, consider using `to_json_schema` and `validate_json`.

    Parameters
    ----------
    data (D | dict):
        The data to be validated, either as a dictionary or an instance of the schema dataclass.
    schema (type[D]):
        The dataclass type representing the schema to validate against.

    Returns
    -------
    D
        An instance of the schema dataclass populated with the validated data.

    Raises
    ------
    ConfigurationError: If the data does not comply with the schema,
        this error is raised with details of the violations.

    """
    json_schema = to_json_schema(schema)
    if is_dataclass(data):
        data = asdict(data)
    validate_json(data, json_schema)
    return dataclass_from_dict(schema, **data)


def validate_json(data: DataclassInstance | dict, schema: dict) -> None:
    """Validate the provided data against the given schema.

    This function uses the Draft202012Validator to check if the data
    conforms to the specified schema. If there are any validation errors,
    it raises a ConfigurationError containing the details of the errors.

    Parameters
    ----------
    data (dict):
        The data to be validated.
    schema (dict):
        The JSON schema to validate against.

    Raises
    ------
    ConfigurationError: If the data does not comply with the schema,
                          this error is raised with details of the violations.
    """
    if is_dataclass(data):
        data = asdict(data)

    schema = allow_none_in_schema(schema)
    validator = Draft202012Validator(schema)  # type: ignore[bad-instantiation]

    errors = list(validator.iter_errors(data))
    if errors:
        import json

        log.error("Validation errors in configuration data!")
        log.debug(f"Data: {json.dumps(data, indent=2)}")
        log.debug(f"Schema: {json.dumps(schema, indent=2)}")
        raise ConfigurationError.from_ValidationErrors(errors)


def allow_none_in_schema(schema: dict | list) -> dict:  # -> dict[Any, Any] | list[Any]:
    """
    Recursively modifies a JSON schema to allow `null` values for all fields.

    This is needed to parse Optional fields that hold dataclasses. May need a revisit later.
    """
    if isinstance(schema, dict):
        # If current schema block has "type"
        if "type" in schema:
            # If the type is a list of types, add 'null' if it's not already present
            if isinstance(schema["type"], list):
                if "null" not in schema["type"]:
                    schema["type"].append("null")
            else:
                # Else, make it a list with the current type and 'null'
                schema["type"] = [schema["type"], "null"]

        # Recursively process properties in objects
        for key, value in schema.items():
            if isinstance(value, (dict, list)):
                allow_none_in_schema(value)

    elif isinstance(schema, list):
        for item in schema:
            allow_none_in_schema(item)

    return cast(dict, schema)


class ConfigurationError(Exception):
    """Exception raised for configuration validation errors."""

    def __init__(self, message="Configuration error", section=None):
        self.message = message
        self.section = section
        super().__init__(self.message)

    def __str__(self):
        """Return the error message."""
        return (
            f"{self.message} in section '{self.section}'"
            if self.section
            else self.message
        )

    @classmethod
    def from_ValidationErrors(cls, error: ValidationError | list[ValidationError]):
        """Create a ConfigurationError from a ValidationError or a list of ValidationErrors."""
        if isinstance(error, list):
            errors = []
            for e in error:
                path = [str(p) for p in e.path]
                errors.append(cls(e.message, ".".join(path)))
            return MultiConfigurationError(errors)
        else:
            return cls(error.message, str(error.path))


class MultiConfigurationError(Exception):
    """Exception raised for multiple configuration validation errors."""

    def __init__(self, errors: list[ConfigurationError]):
        self.errors = errors
        super().__init__("Multiple configuration errors")

    def __str__(self):
        """Return the error message."""
        return "\n".join([str(e) for e in self.errors])



"""Configuration base classes for EYconf.

Allows to create configuration classes based on dataclass schemas,
with validation and serialization capabilities.
"""

from __future__ import annotations

import json
import logging
from copy import deepcopy
from dataclasses import asdict, is_dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    cast,
)

from eyconf.generate_yaml import dataclass_to_yaml
from eyconf.type_utils import get_type_hints_resolve_namespace
from eyconf.utils import (
    AttributeDict,
    dataclass_from_dict,
)
from eyconf.validation import to_json_schema, validate_json

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

# Needs the string escaping to work at runtime as _typeshed is not a real module
D = TypeVar("D", bound="DataclassInstance")

log = logging.getLogger(__name__)


class EYConfBase(Generic[D]):
    """Base class for EYconf.

    Can be used to create custom configuration classes in memory without file I/O.
    """

    _schema: type[D]
    _data: D
    _json_schema: dict
    allow_additional_properties: bool

    def __init__(
        self,
        data: dict | D,
        schema: type[D] | None = None,
        allow_additional_properties: bool = False,
    ):
        if schema is not None:
            self._schema = schema
        else:
            if not is_dataclass(data):
                raise ValueError(
                    "If no schema is provided, data must be of the schema dataclass instance."
                )
            self._schema = type(data)

        # Create schema, raise if Schema is invalid
        self.allow_additional_properties = allow_additional_properties
        self._json_schema = to_json_schema(
            self._schema,
            allow_additional=allow_additional_properties,
        )

        # Will raise ConfigurationError if the data does not comply with the schema
        validate_json(data, self._json_schema)
        if is_dataclass(data):
            self._data = cast(D, data)
        else:
            self._data = self._schema()  # type: ignore[bad-assignment]
            self.update(data)

    def validate(self):
        """Validate the current data against the schema."""
        validate_json(self._data, self._json_schema)

    def update(self, data: dict):
        """Update the configuration with provided data.

        This applies an partial update to the existing configuration data.
        Only the provided keys will be updated, others will remain unchanged.
        """

        def _update(
            target_type: type[DataclassInstance],
            target: DataclassInstance,
            update_data: dict,
            _current_path: list[str] = [],
        ):
            target_annotations = get_type_hints_resolve_namespace(target_type)
            for key, value in update_data.items():
                if hasattr(target, key):
                    # folders : dict[str, InboxFolder]
                    current_value = getattr(target, key)
                    # current_value = {placeholder:Config42()}

                    # Handle dataclass fields
                    if is_dataclass(current_value):
                        _update(
                            target_annotations[key],
                            current_value,  # type: ignore[arg-type]
                            value,
                            _current_path + [key],
                        )
                    # Handle Optional fields that were previously None
                    elif current_value is None and isinstance(value, dict):
                        nested_instance = dataclass_from_dict(
                            target_annotations[key], value
                        )
                        setattr(target, key, nested_instance)
                    elif current_annotation := target_annotations.get(key):
                        nested = dataclass_from_dict(
                            current_annotation,
                            # TODO: If we want to implement a merge strategy
                            # some like this could work:
                            # merge_dicts(asdict(current_value), value),  # type: ignore[arg-type]
                            value,
                        )
                        setattr(target, key, nested)
                    else:
                        # Primitives and direct assignments
                        setattr(target, key, value)
                else:
                    # Non-schema fields (EYConfAdditional)
                    self._update_additional(
                        target, key, value, _current_path=_current_path
                    )

        old_data = deepcopy(self._data)
        update_dict = data if not is_dataclass(data) else asdict(data)

        _update(self._schema, self._data, update_dict)

        try:
            self.validate()
        except Exception as e:
            self._data = old_data
            raise e from e

    def _update_additional(self, target, key, value: Any, _current_path: list[str]):
        if not self.allow_additional_properties:
            raise AttributeError(
                "Cannot set unknown attribute"
                f" '{'.'.join(_current_path + [key])}' on configuration."
            )
        if isinstance(value, dict):
            setattr(target, key, AttributeDict(**value))
        else:
            setattr(target, key, value)

    def overwrite(self, data: dict | D):
        """Overwrite the configuration with provided data.

        If the provided data is missing required fields, an error will be raised.
        """
        data = asdict(data) if is_dataclass(data) else data
        validate_json(data, self._json_schema)
        self._data = dataclass_from_dict(self._schema, data)

    def reset(self):
        """Reset the configuration data to the default values."""
        self._data = self._schema()

    # -------------------------------- Converters -------------------------------- #

    @property
    def data(self) -> D:
        """Get the configuration data."""
        return self._data

    def default_yaml(self) -> str:
        """Return the configs' defaults (inferred from schema) as yaml.

        You may overwrite this method to customize the default configuration
        generation.
        """
        return dataclass_to_yaml(self._schema)

    def to_dict(self, include_additional: bool = False) -> dict:
        """Convert the configuration data to a dictionary.

        Parameters
        ----------
        include_additional : bool
            Whether to include extra data not part of the schema.
        """
        if include_additional:
            # A bit hacky but works for now
            return json.loads(json.dumps(self._data, default=lambda o: o.__dict__))
        return asdict(self._data)

    def to_yaml(self) -> str:
        """Convert the configuration data to a yaml string."""
        return dataclass_to_yaml(self._data)

    # --------------------------------- Printing --------------------------------- #

    def __repr__(self) -> str:
        """Return a custom string representation of the configuration object."""
        class_name = type(self).__name__
        memory_address = hex(id(self))
        prefix = f"<{class_name} object at {memory_address}>:\n"
        return f"{prefix}{self.__str__()}"

    def __str__(self):
        """Return a custom formatted string representation of the configuration data."""
        data_dict = asdict(self._data)
        s = "\n".join(
            "  " + line for line in self._pretty_format(data_dict).splitlines()
        )
        return s

    def _pretty_format(self, data, indent=0):
        """Format the dict with pretty indentation."""
        result: list[str] = []
        for key, value in data.items():
            if isinstance(value, dict):
                result.append(" " * indent + f"{key}:")
                result.append(self._pretty_format(value, indent + 4))
            else:
                result.append(" " * indent + f"{key}: {value}")
        return "\n".join(result)



from __future__ import annotations

from dataclasses import dataclass
import os
from pathlib import Path
from unittest.mock import patch
import pytest
from eyconf import EYConf

try:
    from typer.testing import CliRunner
    from eyconf.cli import create_config_cli

    typer_installed = True
except ImportError:
    typer_installed = False


skip = pytest.mark.skipif(
    not typer_installed, reason="typer is not installed, skipping CLI tests."
)


@dataclass
class Config:
    """Example configuration data class."""

    int_field: int = 42
    str_field: str = "Hello, World!"


@pytest.fixture(autouse=True)
def mock_get_file_path(tmp_path) -> Path:
    """Fixture to provide a temporary config file path."""
    config_file_path = tmp_path / "config.yml"
    os.environ["EYCONF_CONFIG_FILE"] = str(config_file_path)
    return config_file_path


@pytest.fixture
@skip
def cli_app():
    """Fixture to create a CLI app for the configuration commands."""
    config_cli = create_config_cli(EYConf, schema=Config)
    return config_cli


@skip
def test_main_command_shows_help(cli_app, mock_get_file_path):
    """Test that invoking the main command without subcommands shows help."""
    runner = CliRunner()
    result = runner.invoke(cli_app, [])

    assert result.exit_code == 0
    assert "Manage configuration file" in result.output


@skip
def test_ls_command(cli_app, mock_get_file_path):
    """Test the 'ls' command to list current configuration."""
    runner = CliRunner()

    # Instantiate the config to ensure file exists.
    EYConf(Config)

    result = runner.invoke(cli_app, ["ls"])

    assert result.exit_code == 0
    assert "int_field" in result.output
    assert "str_field" in result.output


@skip
def test_edit_command(monkeypatch, cli_app, mock_get_file_path):
    """Test the 'edit' command to open the configuration file in an editor."""

    async def mock_asyncio_create_subprocess_exec(*args, **kwargs):
        """Mock subprocess execution for opening a file."""

        class MockProcess:
            async def wait(self):
                pass

        return MockProcess()

    with patch("asyncio.create_subprocess_exec", mock_asyncio_create_subprocess_exec):
        runner = CliRunner()
        result = runner.invoke(cli_app, ["--edit"])

    print(result.output)  # For debugging purposes

    assert result.exit_code == 0
    assert "Opening configuration file:" in result.output



from dataclasses import dataclass, field
import pytest
from eyconf.config import EYConfExtraFields
from eyconf.utils import AccessProxy


@dataclass
class Config42:
    int_field: int = 42
    str_field: str = "FortyTwo!"


class TestEYConfExtraFields:
    def test_init(self):
        config = EYConfExtraFields(Config42())

        assert isinstance(config.data, AccessProxy)
        assert config.data.int_field == 42

    def test_set_data(self):
        config = EYConfExtraFields(Config42())

        config.data.int_field = 100
        config.data.new_field = "Hundred"  # type: ignore

        assert config.data.int_field == 100
        assert config.data.new_field == "Hundred"  # type: ignore

        assert config.to_dict() == {
            "int_field": 100,
            "str_field": "FortyTwo!",
            "new_field": "Hundred",
        }

        assert config.to_dict(False) == {
            "int_field": 100,
            "str_field": "FortyTwo!",
        }

    def test_set_mixed(self):
        config = EYConfExtraFields(Config42())

        config.data.new_field = "New Value"  # type: ignore
        config.data.nested.foo = "Bar"  # type: ignore
        config.data.nested.very_deep.another_level = 123  # type: ignore

        assert config.data.new_field == "New Value"  # type: ignore
        assert config.data.nested.foo == "Bar"  # type: ignore
        assert config.data.nested.very_deep.another_level == 123  # type: ignore

        assert config.to_dict() == {
            "int_field": 42,
            "str_field": "FortyTwo!",
            "new_field": "New Value",
            "nested": {
                "foo": "Bar",
                "very_deep": {
                    "another_level": 123,
                },
            },
        }

        assert config.to_dict(False) == {
            "int_field": 42,
            "str_field": "FortyTwo!",
        }

    def test_update_unknown_field(self):
        config = EYConfExtraFields(Config42())

        config.update({"unknown_field": "I am unknown!"})

        assert config.data.unknown_field == "I am unknown!"  # type: ignore[attr-defined]
        assert config._extra_data.unknown_field == "I am unknown!"
        with pytest.raises(AttributeError):
            _ = config._data.non_existent_field  # type: ignore[attr-defined]

    def test_update_dict(self):
        @dataclass
        class SchemaDict:
            folders: dict[str, Config42] = field(
                default_factory=lambda: {"placeholder": Config42()}
            )

        config = EYConfExtraFields(SchemaDict())

        assert config.data.folders["placeholder"].int_field == 42
        for folder in config.data.folders.values():
            assert isinstance(folder, Config42)

        config.update(
            {
                "folders": {
                    "config1": {"int_field": 1, "str_field": "One"},
                    "config2": {"int_field": 2, "str_field": "Two"},
                }
            }
        )

        assert config.data.folders["config1"].int_field == 1
        assert config.data.folders["config1"].str_field == "One"
        assert config.data.folders["config2"].int_field == 2
        assert config.data.folders["config2"].str_field == "Two"

        # We should revise this with a merge strategy later
        with pytest.raises(KeyError):
            config.data.folders["placeholder"]

    def test_update_dict_nested(self):
        @dataclass
        class NestedSchema:
            folders: dict[str, Config42] = field(
                default_factory=lambda: {"placeholder": Config42()}
            )

        @dataclass
        class SchemaWithNested:
            nested: NestedSchema = field(default_factory=NestedSchema)

        config = EYConfExtraFields(SchemaWithNested())

        assert config.data.nested.folders["placeholder"].int_field == 42
        for folder in config.data.nested.folders.values():
            assert isinstance(folder, Config42)

        config.update(
            {
                "nested": {
                    "folders": {
                        "config1": {"int_field": 1, "str_field": "One"},
                        "config2": {"int_field": 2, "str_field": "Two"},
                    }
                }
            }
        )

        assert config.data.nested.folders["config1"].int_field == 1
        assert config.data.nested.folders["config1"].str_field == "One"
        assert config.data.nested.folders["config2"].int_field == 2
        assert config.data.nested.folders["config2"].str_field == "Two"

    def test_update_unknown_nested(self):
        config = EYConfExtraFields(Config42())

        config.update({"level_one": {"level_two": {"level_three": "Deep Value"}}})
        assert config.data.level_one.level_two.level_three == "Deep Value"  # type: ignore[attr-defined]



from dataclasses import asdict
from typing import TYPE_CHECKING, Any, TypeVar

from eyconf.utils import AccessProxy, AttributeDict, merge_dicts

from .base import EYConfBase

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

# Needs the string escaping to work at runtime as _typeshed is not a real module
D = TypeVar("D", bound="DataclassInstance")


class EYConfExtraFields(EYConfBase[D]):
    """Configuration class that supports extra fields explicitly.

    This class extends the base configuration functionality to allow
    for additional fields that are not defined in the original schema.

    This additional field is accessible via the `extra_data_as_dict` property.
    """

    _extra_data: AttributeDict

    def __init__(
        self,
        data: dict | D,
        schema: type[D] | None = None,
    ):
        super().__init__(data, schema=schema, allow_additional_properties=True)
        self._extra_data = AttributeDict()

    @property
    def data(self) -> D:
        """Get the configuration data wrapped in a dynamic accessor.

        Care: Instance checks will not work as expected on this property.
        """
        return AccessProxy(self._data, self._extra_data)  # type: ignore

    @property
    def extra_data(self) -> AttributeDict:
        """Get the extra data as an AttributeDict."""
        return self._extra_data

    def to_dict(self, include_additional: bool = True) -> dict:
        """Get the full configuration data as a dictionary, including extra fields."""
        data = asdict(self._data)
        if include_additional:
            data = merge_dicts(data, self.extra_data.as_dict())
        return data

    def _update_additional(self, target, key, value: Any, _current_path: list[str]):
        extra_data: AttributeDict = self._extra_data
        for path_part in _current_path:
            extra_data = getattr(extra_data, path_part)

        if isinstance(value, dict):
            setattr(extra_data, key, AttributeDict(**value))
        else:
            setattr(extra_data, key, value)



import inspect
import logging
import sys
from dataclasses import is_dataclass
from typing import (
    Any,
    get_type_hints,
)

log = logging.getLogger(__name__)


def get_type_hints_resolve_namespace(obj, include_extras: bool = False):
    """Get type hints for an object, resolving namespaces for dataclasses.

    Workaround for when using `from __future__ import annotations`.

    This is needed because Types become strings, and native
    recursive resolution does not work on strings alone.
    To resolve types recursively in this case, the Dataclasses are needed
    and can be passed to `get_type_hints` via `globalns` and `localns`.
    """
    try:
        return get_type_hints(obj, include_extras=include_extras)
    except NameError:
        globalns, localns = _get_namespace(obj)
        return get_type_hints(
            obj,
            globalns=globalns,
            localns=localns,
            include_extras=include_extras,
        )


def _get_namespace(
    obj,
) -> tuple[dict[str, Any], dict[str, Any]]:
    """Get the global and local namespaces for a dataclass.

    I.e. where to find the dataclass definitions and possibly
    other types used in the dataclass.

    Workaround for when using `from __future__ import annotations`.

    This is needed because Types become strings, and native
    recursive resolution does not work on strings alone.
    To resolve types recursively in this case, the Dataclasses are needed
    and can be passed to `get_type_hints` via `globalns` and `localns`.
    """
    module = sys.modules.get(obj.__module__)
    globalns = getattr(module, "__dict__", {}) if module else {}

    # Get the caller's locals to handle locally defined classes
    frame = inspect.currentframe()
    localns = {}

    # Walk up the call stack to find local variables
    while frame:
        frame_locals = frame.f_locals

        # Merge locals that might contain our dataclass definitions
        for name, value in frame_locals.items():
            if inspect.isclass(value):
                localns[name] = value
                globalns[name] = value  # Also add to globals

        frame = frame.f_back

    return globalns, localns


def is_dataclass_type(obj: Any) -> bool:
    """Check if an object is a dataclass type (class), not an instance."""
    return is_dataclass(obj) and isinstance(obj, type)



"""Default file-based configuration class.

Allows to generate, validate and load a yaml configuration file based
on a dataclass schema.
"""

from __future__ import annotations

import logging
import os
from dataclasses import is_dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    TypeVar,
)

import yaml

from eyconf.utils import (
    dataclass_from_dict,
)
from eyconf.validation import to_json_schema, validate_json

from .base import EYConfBase

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

# Needs the string escaping to work at runtime as _typeshed is not a real module
D = TypeVar("D", bound="DataclassInstance")

log = logging.getLogger(__name__)


class EYConf(EYConfBase[D]):
    """Configuration class.

    This class is used to generate a default configuration file from a schema
    represented by a dataclass.

    It allows to generate, validate and load a configuration file.
    """

    path: Path

    def __init__(
        self,
        schema: type[D],
    ):
        if not is_dataclass(schema) or not isinstance(schema, type):
            raise ValueError(
                "Schema must be a dataclass class. Instances are not supported yet."
            )
        self.path = self.get_file()

        # Generate default configuration if it does not exist
        self._schema = schema
        self._json_schema = to_json_schema(self._schema)
        if not self.path.exists():
            try:
                self._data = schema()
            except TypeError:
                log.exception(
                    "Schema dataclass has required fields without defaults. Consider using field with default_factory in your schema."
                )
                raise
            self._write_default()
        else:
            self._data = self._load_and_validate()

    @staticmethod
    def get_file() -> Path:
        """Get the path to the configuration file."""
        return (
            Path(os.environ.get("EYCONF_CONFIG_FILE", "./config.yaml"))
            .expanduser()
            .resolve()
        )

    def reset(self):
        """Reset the configuration by reloading and validating the file."""
        self._data = self._load_and_validate()

    def __repr__(self) -> str:
        """Return a custom string representation of the configuration object."""
        class_name = type(self).__name__
        memory_address = hex(id(self))
        prefix = f"<{class_name} object at {memory_address} loaded from {self.path.absolute()}>:\n"

        return f"{prefix}{self.__str__()}"

    # ------------------ Helpers for file generation and loading ----------------- #

    def _write_default(self):
        """Generate default yaml configuration."""
        if self.path.exists():
            log.warning(f"Configuration file {self.path} already exists. Overwriting!")

        yaml_str = self.default_yaml()
        os.makedirs(self.path.parent, exist_ok=True)
        with open(self.path, "w") as f:
            f.write(yaml_str)
            f.write("\n")  # Add a newline at the end of the file
        log.info(f"Configuration file created at '{self.path.absolute()}'")

    def _load_and_validate(self) -> D:
        """Load the configuration file and validate it against the schema."""
        log.info(f"Loading config file: {self.path.absolute()}")

        if not self.path.exists():
            raise FileNotFoundError(
                f"Configuration file '{self.path.absolute()}' not found. Please generate with `write_default()`."
            )

        # Load the config file
        with open(self.path) as file:
            data = yaml.safe_load(file)

        # Will raise ConfigurationError if the data does not comply with the schema
        validate_json(data, self._json_schema)

        return dataclass_from_dict(self._schema, data)
